# -*- coding: utf-8 -*-
from typing import Iterable, Set, Optional, List, Iterator, Any, Dict

import re
import csv
import sys
import datetime
import logging

from ofxstatement.plugin import Plugin as BasePlugin
from ofxstatement.parser import CsvStatementParser
from ofxstatement.exceptions import ParseError
from ofxstatement.statement import BankAccount

from ofxstatement.plugins.nl.statement import Statement, StatementLine

# Need Python 3 for super() syntax
assert sys.version_info[0] >= 3, "At least Python 3 is required."

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Parser(CsvStatementParser):
    """

    These are the first two lines of an ING Netherlands CSV file:

    "Datum","Naam / Omschrijving","Rekening","Tegenrekening","Code",\
"Af Bij","Bedrag (EUR)","MutatieSoort",\
"Mededelingen"
    "20200213","Kosten OranjePakket met korting","NL42INGB0001085276","","DV",\
"Af","1,25","Diversen",\
"1 jan t/m 31 jan 2020 ING BANK N.V. Valutadatum: 13-02-2020"

    These fields are from the Statement class:

    id = ""

    # Date transaction was posted to account (booking date)
    date = datetime.now()

    memo = ""

    # Amount of transaction
    amount = D(0)

    # additional fields
    payee = ""

    # Date user initiated transaction, if known (transaction date)
    date_user = datetime.now()

    # Check (or other reference) number
    check_no = ""

    # Reference number that uniquely identifies the transaction. Can be used in
    # addition to or instead of a check_no
    refnum = ""

    # Transaction type, must be one of TRANSACTION_TYPES
    "CREDIT",       # Generic credit
    "DEBIT",        # Generic debit
    "INT",          # Interest earned or paid
    "DIV",          # Dividend
    "FEE",          # FI fee
    "SRVCHG",       # Service charge
    "DEP",          # Deposit
    "ATM",          # ATM debit or credit
    "POS",          # Point of sale debit or credit
    "XFER",         # Transfer
    "CHECK",        # Check
    "PAYMENT",      # Electronic payment
    "CASH",         # Cash withdrawal
    "DIRECTDEP",    # Direct deposit
    "DIRECTDEBIT",  # Merchant initiated debit
    "REPEATPMT",    # Repeating payment/standing order
    "OTHER"         # Other

    trntype = "CHECK"

    # Optional BankAccount instance
    bank_account_to = None

    """

    date_format: str

    # transactions / balance
    header: List[List[str]] = [["Datum",
                                "Naam / Omschrijving",
                                "Rekening",
                                "Tegenrekening",
                                "Code",
                                "Af Bij",
                                "Bedrag (EUR)",
                                "MutatieSoort",
                                "Mededelingen"],
                               ["Datum",
                                "Boeksaldo",
                                "Valutair saldo"]]
    # 0-based
    mappings_by_header: List[Dict[str, int]] = [{
        # id (determined later)
        'date': 0,
        'memo': 8,
        'amount': 6,
        'payee': 1,  # if bank_account_to is filled
        # date_user
        # check_no
        # refnum
        # trntype (determined later)
        'bank_account_to': 3,
    }, {
        'date': 0,
        'amount': 2  # valutair
    }]

    # variables
    unique_id_set: Set[str]
    header_idx: int
    mappings: Dict[str, int]

    def __init__(self,
                 fin: Iterable[str],
                 account_id: Optional[str] = None) -> None:
        # Python 3 needed
        super().__init__(fin)
        # Use the BIC code for ING Netherlands
        self.statement = Statement(bank_id="INGBNL2A",
                                   account_id=account_id,
                                   currency="EUR")  # My Statement
        self.unique_id_set = set()
        self.header_idx = -1

    def parse(self) -> Statement:
        """Main entry point for parsers

        super() implementation will call to split_records and parse_record to
        process the file.
        """

        # Python 3 needed
        stmt: Statement = super().parse()

        try:
            assert self.header_idx >= 0 and self.header_idx < len(self.header),\
                "Header not read: {}".format(str(self.header))
        except Exception as e:
            raise ParseError(0, str(e))

        if self.header_idx == 0:
            # GJP 2020-03-03
            # No need to (re)calculate the balance since there is no history.
            # But set the dates.
            stmt.start_balance = stmt.end_balance = None
            stmt.start_date = min(sl.date for sl in stmt.lines)
            # end date is exclusive for OFX
            stmt.end_date = max(sl.date for sl in stmt.lines)
            stmt.end_date += datetime.timedelta(days=1)
        elif self.header_idx == 1:
            stmt.start_date = stmt.start_balance = None

            stmt.end_date = max(sl.date for sl in stmt.lines)
            assert stmt.lines[0].date == stmt.end_date or \
                stmt.lines[-1].date == stmt.end_date
            end_idx: int = 0 if stmt.lines[0].date == stmt.end_date else -1
            stmt.end_balance = stmt.lines[end_idx].amount
            # end date is exclusive for OFX
            stmt.end_date += datetime.timedelta(days=1)

            # no transaction lines
            stmt.lines = []

        return stmt

    def split_records(self) -> Iterator[Any]:
        """Return iterable object consisting of a line per transaction
        """
        return csv.reader(self.fin, delimiter=',')

    def parse_record(self,
                     line: List[Optional[str]]) -> Optional[StatementLine]:
        """Parse given transaction line and return StatementLine object
        """

        stmt_line: Optional[StatementLine] = None

        try:
            logger.debug('header idx: %d; line #%d: %s',
                         self.header_idx,
                         self.cur_record,
                         line)

            # First record(s) must be the header
            if self.header_idx < 0:
                if line == self.header[0]:
                    self.header_idx = 0
                    self.date_format = "%Y%m%d"
                elif line == self.header[1]:
                    self.header_idx = 1
                    self.date_format = "%Y-%m-%d"

                msg: str = "Line {} does not match\n\n{}\n\nnor\n\n{}\n"

                assert self.header_idx in [0, 1],\
                    msg.format(line, self.header[0], self.header[1])

                self.mappings = self.mappings_by_header[self.header_idx]
                return None

            if self.header_idx == 0:
                stmt_line = self.parse_transaction(line)
            elif self.header_idx == 1:
                stmt_line = self.parse_balance(line)

        except Exception as e:
            raise ParseError(self.cur_record, str(e))

        return stmt_line

    def parse_transaction(self,
                          line: List[Optional[str]]) -> Optional[StatementLine]:
        # line[2] contains the account number
        if self.statement.account_id:
            assert self.statement.account_id == line[2],\
                "Only one account is allowed; previous account: {}, \
this line's account: {}".format(self.statement.account_id, line[2])
        else:
            self.statement.account_id = line[2]

        assert line[5] in ['Af', 'Bij']

        if line[5] == 'Af':
            amount: Optional[str] = line[self.mappings['amount']]
            line[self.mappings['amount']] =\
                '-' + (amount if isinstance(amount, str) else '')

        if line[self.mappings['bank_account_to']]:
            line[self.mappings['payee']] =\
                "{} ({})".format(line[self.mappings['payee']],
                                 line[self.mappings['bank_account_to']])
        else:
            line[self.mappings['memo']] =\
                "{}, {}".format(line[self.mappings['payee']],
                                line[self.mappings['memo']])
            line[self.mappings['payee']] = None

        # Python 3 needed
        stmt_line: StatementLine = super().parse_record(line)

        # Remove zero-value notifications
        if stmt_line.amount == 0:
            return None

        # Determine some fields not in the self.mappings
        # A hack but needed to use the adjust method
        stmt_line.__class__ = StatementLine
        stmt_line.adjust(self.unique_id_set)

        if stmt_line.amount < 0:
            stmt_line.trntype = "DEBIT"
        else:
            stmt_line.trntype = "CREDIT"

        if stmt_line.bank_account_to:
            stmt_line.bank_account_to = \
                BankAccount(bank_id=None,
                            acct_id=stmt_line.bank_account_to)
        return stmt_line

    def parse_balance(self,
                      line: List[Optional[str]]) -> Optional[StatementLine]:
        # Python 3 needed
        stmt_line: StatementLine = super().parse_record(line)
        stmt_line.trntype = "DEBIT" if stmt_line.amount < 0 else "CREDIT"
        stmt_line.id = 1
        # Determine some fields not in the self.mappings
        # A hack but needed to use the adjust method
        stmt_line.__class__ = StatementLine
        stmt_line.adjust(self.unique_id_set)
        return stmt_line


class Plugin(BasePlugin):
    """ING Bank, The Netherlands, CSV (https://www.ing.nl/)
    """
    def get_parser(self, filename: str) -> Parser:
        p = re.compile('(NL\\d+INGB\\d+)')
        m = p.search(filename)
        account_id: Optional[str] = None
        if m:
            account_id = m.group(0)
        fin = open(filename, "r", encoding="ISO-8859-1")
        return Parser(fin, account_id)



# -*- coding: utf-8 -*-
from typing import Iterable, Set, Optional, List, Iterator, Any

import sys
import locale
import re
import io
from decimal import Decimal
from datetime import datetime, date as dt
from subprocess import check_output, CalledProcessError
import logging

from ofxstatement.plugin import Plugin as BasePlugin
from ofxstatement.parser import StatementParser as BaseStatementParser

from ofxstatement.plugins.nl.statement import Statement, StatementLine

# Need Python 3 for super() syntax
assert sys.version_info[0] >= 3, "At least Python 3 is required."

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Parser(BaseStatementParser):
    unique_id_set: Set[str]

    def __init__(self, fin: Iterable[str]) -> None:
        super().__init__()
        self.statement = Statement(bank_id=None,
                                   account_id=None,
                                   currency='EUR')  # My Statement
        self.fin = fin
        self.unique_id_set = set()

    def parse(self) -> Optional[Statement]:
        """Main entry point for parsers

        super() implementation will call to split_records and parse_record to
        process the file.
        """
        stmt: Optional[Statement] = None
        # Save locale
        current_locale = locale.setlocale(category=locale.LC_ALL)
        # Need to parse "05 mei" i.e. "05 may"
        locale.setlocale(category=locale.LC_ALL, locale="nl_NL")
        try:
            # Python 3 needed
            stmt = super().parse()

            if stmt and stmt.lines:
                stmt.start_date = min(sl.date for sl in stmt.lines)
        finally:
            locale.setlocale(category=locale.LC_ALL, locale=current_locale)

        return stmt

    @staticmethod
    def get_amount(amount_in: str, transaction_type_in: str) -> Decimal:
        sign_out: int = 1
        amount_out: Any

        # determine sign_out
        assert isinstance(transaction_type_in, str)
        assert transaction_type_in in ['Af', 'Bij']

        if transaction_type_in == 'Af':
            sign_out = -1

        # determine amount_out
        assert isinstance(amount_in, str)
        # Amount something like 1.827,97, € 1.827,97 (both dutch) or 1,827.97?
        m = re.search(r'^(\S+\s)?([0-9,.]+)$', amount_in)
        assert m is not None
        amount_out = m.group(2)
        if amount_out[-3] == ',':
            amount_out = amount_out.replace('.', '').replace(',', '.')

        # convert to str to keep just the last two decimals
        return sign_out * Decimal(str(amount_out))

    def split_records(self) -> Iterator[Any]:
        """Return iterable object consisting of a line per transaction
        """
        def convert_str_to_list(str: str,
                                max_items: Optional[int] = None,
                                sep: str = r'\s\s+|\t|\n') -> List[str]:
            return [x for x in re.split(sep, str)[0:max_items]]

        first_line = True
        first_line_row = ['International Card Services BV', 'www.icscards.nl']

        new_page = False
        new_page_row = ['Datum', 'ICS-klantnummer', 'Volgnummer', 'Bladnummer']

        balance = False
        balance_row = ['Vorig openstaand saldo', 'Totaal ontvangen betalingen',
                       'Totaal nieuwe uitgaven', 'Nieuw openstaand saldo']

        statement_expr = \
            re.compile(r'^\d\d [a-z]{3}\s+\d\d [a-z]{3}.+[0-9,.]+\s+(Af|Bij)$')

        for line in self.fin:
            line = line.strip()

            logger.debug('line: %s', line)

            # to ease the parsing pain
            row: List[str] = convert_str_to_list(line)

            if first_line and len(row) > 1:
                assert row == first_line_row,\
                    "Expected: {0}\nActual: {1}".format(first_line_row, row)
                first_line = False

            if len(row) == 2 and row[1][0:5] == 'BIC: ':
                self.statement.bank_id = row[1][5:]

            elif row == new_page_row:
                new_page = True
            elif new_page:
                new_page = False
                self.statement.end_date = row[0]  # exclusive in ICSCards
                self.statement.end_date = \
                    datetime.strptime(self.statement.end_date,
                                      '%d %B %Y').date()
                self.statement.account_id = row[1]

            elif row == balance_row:
                balance = True
            elif balance:
                balance = False
                self.statement.start_balance = Parser.get_amount(row[0],
                                                                 row[1])
                self.statement.end_balance = Parser.get_amount(row[-2],
                                                               row[-1])

            elif re.search(statement_expr, line):
                # payee (column 2), place and contry may be something like:
                #
                # THY|2357312380512|Istanbul|US
                #
                # hence four columns instead of three, so combine the first two
                country = re.compile("^[A-Z][A-Z]$")
                for i in reversed(range(len(row))):
                    if country.match(row[i]):
                        # Should have 4 columns to the left. If not: reduce.
                        while i > 4:
                            row[2] += ' ' + row[3]
                            del row[3]
                            i -= 1
                        break

                if len(row) >= 6 and len(row) <= 7 and len(row[2]) > 25:
                    # 04 sep | 05 sep | NEWREST WAGONS LITS FRANCPARIS ...
                    # =>
                    # 04 sep | 05 sep | NEWREST WAGONS LITS FRANC ...

                    row.insert(2, row[2][0:25])
                    row[3] = row[3][25:]
                yield row

    def parse_record(self, row: List[str]) -> Optional[StatementLine]:
        """Parse given transaction line and return StatementLine object
        """

        def add_years(d: dt, years: int) -> dt:
            """Return a date that's `years` years after the date (or datetime)
            object `d`. Return the same calendar date (month and day) in the
            destination year, if it exists, otherwise use the following day
            (thus changing February 29 to March 1).

            """
            return d.replace(year=d.year + years, month=3, day=1) \
                if d.month == 2 and d.day == 29 \
                else d.replace(year=d.year + years)

        def get_date(d_m: str) -> dt:
            # Without a year it will be 1900 so add the year
            d_m_y = "{} {}".format(d_m, self.statement.end_date.year)
            d = datetime.strptime(d_m_y, '%d %b %Y').date()
            # But now the resulting date may be more than the end date
            # (d_m in december and end date in january)
            if d > self.statement.end_date:
                d = add_years(d, -1)
            assert d <= self.statement.end_date
            return d

        logger.debug('row: %s', str(row))
        assert(len(row) in [5, 7, 8])

        stmt_line: Optional[StatementLine] = None
        # GJP 2020-03-01
        # Skip transaction date (index 0) since it gives a wrong balance.
        # Use booking date (index 1) in order to get a correct balance.
        date = get_date(row[1])

        payee = None
        memo = None
        if len(row) >= 7:
            # payee (2), place (3) and country (4)
            payee = row[2]
            memo = "{0} ({1})".format(row[3], row[4])
        else:
            # description (2)
            memo = row[2]

        # Skip amount in foreign currency
        amount = Parser.get_amount(row[-2], row[-1])

        # Remove zero-value notifications
        if amount != 0:
            stmt_line = StatementLine(date=date,
                                      memo=memo,
                                      amount=amount)
            stmt_line.payee = payee
            stmt_line.adjust(self.unique_id_set)

        return stmt_line


class Plugin(BasePlugin):
    """ICSCards, The Netherlands, PDF (https://icscards.nl/)
    """

    def get_file_object_parser(self, fh: Iterable[str]) -> Parser:
        return Parser(fh)

    def get_parser(self, filename: str) -> Parser:
        pdftotext = ["pdftotext", "-layout", filename, '-']
        fh: Iterable[str]

        # Is it a PDF or an already converted file?
        try:
            fh = io.StringIO(check_output(pdftotext).decode())
            # No exception: apparently it is a PDF.
        except CalledProcessError:
            fh = open(filename, "r")

        return self.get_file_object_parser(fh)
